/*
This file is part of libmspcore     
Copyright © 2006  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#ifndef MSP_TIME_TIMER_H_
#define MSP_TIME_TIMER_H_

#include <queue>
#include <sigc++/sigc++.h>
#include "../core/mutex.h"
#include "../core/semaphore.h"
#include "timedelta.h"
#include "timestamp.h"

namespace Msp {
namespace Time {

/**
A class for executing functions in a deferred or periodical fashion.  The add a
timer, use one of the add functions and connect a functor to the timeout signal
of the returned slot.

This class is thread-safe, to allow running timers in a separate thread.
*/
class Timer
{
public:
	class Slot
	{
	public:
		sigc::signal<bool> signal_timeout;
	
	private:
		TimeDelta interval;
		TimeStamp timeout;
	
	public:
		Slot(const TimeDelta &);
		Slot(const TimeStamp &);
		const TimeStamp &get_timeout() const { return timeout; }
		bool increment();
		bool operator<(const Slot &) const;
	};

private:
	typedef bool (*fSlotCompare)(Slot *, Slot *);

	std::priority_queue<Slot *, std::vector<Slot *>, fSlotCompare> slots;
	Semaphore sem;
	Mutex mutex;

public:
	Timer();
	~Timer();

	/**
	Adds a timer that will be executed periodically as long as the timeout
	signal hander returns true.
	*/
	Slot &add(const TimeDelta &);

	/**
	Adds a timer that will be executed once at a specific time.  The return
	value of the timeout signal handler is ignored.
	*/
	Slot &add(const TimeStamp &);

	/**
	Checks all timers, executing any that have timed out.  If block is true,
	waits until one times out.

	Note: If there are no active timers when a blocking tick is executed, it
	won't return until a timer is added from another thread.
	*/
	void tick(bool block=true);
private:
	static bool slot_compare(Slot *, Slot *);
};

} // namespace Time
} // namespace Msp

#endif
